/*
 * Copyright (c) 2007, Gregory Fleischer (gfleischer@gmail.com)
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the author nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
/*
 * $Id: find_md5_prefix_collision.c 4 2007-04-07 03:45:32Z gregory $
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <getopt.h>
#include <sys/mman.h>
#include "md5.h"

#define MD5BUFLEN 16

#if !defined(countof)
# define countof(a) (sizeof(a)/sizeof(a[0]))
#endif

const char *progname;
const char *filename;
int verbose = 0;
int badchars = 0;

void
usage()
{
  printf("\n");
  printf("Usage: %s [options] <file>\n", progname);
  printf("  -h, --help       print this usage and exit\n");
  printf("  -v, --verbose    be more verbose\n");
  printf("  -b, --badchars   check for bad characters\n");
  printf("  -t, --tunnel     use Vlastimil Klima's md5_vk1_single\n");
  printf("  -s, --seed <val> use 'val' as seed to md5_vk1_single\n");
  printf("\n");
  fflush(stdout);
}

void
die(int exitcode, const char *fmt, ...)
{
  va_list va;

  fprintf(stderr, "error: ");
  va_start(va, fmt);
  vfprintf(stderr, fmt, va);
  va_end(va);
  putc('\n', stderr);
  fflush(stderr);
  if (exitcode > 0) {
    exit(exitcode);
  }
}

int
check_badchars(const unsigned char *buf, size_t len)
{
  size_t i;
  for (i = 0; i < len; ++i) {
    switch (buf[i]) {
    case 0x00:
    case 0x09:
    case 0x0a:
    case 0x0d:
    case '"':
    case '<':
    case '>':
    case '&':
/*     case 0x80: */
/*     case 0x81: */
/*     case 0x8d: */
/*     case 0x8f: */
/*     case 0x90: */
/*     case 0x9d: */
      return buf[i];
    }
  }

  return -1;
}

int
check_file(const unsigned char *data, size_t datalen, const char *file, unsigned char *signature)
{
  struct MD5Context MD5c;
  struct stat st;
  int i, ch;
  int fd;
  size_t len;
  void *addr;
  unsigned char *buf;

  if (  stat(file, &st) < 0) {
    die(1, "failed to stat %s: %s", file, strerror(errno));
  }
  len = st.st_size;
  if ( (fd = open(file, O_RDONLY)) < 0) {
    die(1, "failed to open %s: %s", file, strerror(errno));
  }
  if ( (addr = mmap(0, len, PROT_READ, MAP_PRIVATE, fd, 0)) == NULL) {
    die(1, "failed to mmap %d: %s", fd, strerror(errno));
  }
  buf = (unsigned char *)addr;

  /* check bad chars if requested */
  if (badchars) {
    if ( (ch = check_badchars(buf, len)) > -1) {
      if (verbose) {
	printf("file %s contains bad character: %c (0x%02x)\n", 
	       file, (isgraph(ch) ? ch : '.'), ch );
      }
      munmap(addr, len);
      close(fd);
      return 0;
    }
  }

  /* now, compute the MD5 value of the prefix + file */
  MD5Init(&MD5c);
  MD5Update(&MD5c, data, datalen);
  MD5Update(&MD5c, buf, len);
  MD5Final(signature, &MD5c);

  if (verbose) {
    printf("computed md5 for %s+%s: ", filename, file);
    for (i = 0; i < MD5BUFLEN; ++i) {
      printf("%02x", signature[i]);
    }
    putchar('\n');
  }

  munmap(addr, len);
  close(fd);
  return 1;

}

int
check_files(const unsigned char *data, size_t datalen, const char *file1, const char *file2)
{
  unsigned char signature1[MD5BUFLEN], signature2[MD5BUFLEN];
  int okay = 0;

  if (verbose) {
    printf("validating collision files: %s and %s\n", file1, file2);
  }

  okay = check_file(data, datalen, file1, signature1);
  okay &= check_file(data, datalen, file2, signature2);

  if (okay) {
    if (0 == memcmp(signature1, signature2, MD5BUFLEN)) {
      if (verbose) {
	printf("md5 values match for both files\n");
      }
      return 1;
    } else {
      if (verbose) {
	printf("*** md5 values DO NOT match for both files ***\n");
      }
    }
  } else {
      if (verbose) {
	printf("file checks failed\n");
      }
  }
  return 0;
}

int
main(int argc, char **argv)
{
  int ch, longind;
  int tunnel = 0;
  char *seed = "DEADBEEF";
  char seedbuf[10];
  struct stat st;
  size_t len;
  int fd, devnullfd;
  unsigned char *data;
  void *addr;
  char IV[4][12];
  struct MD5Context MD5c;
  char junk[MD5BUFLEN];
  int pid, status, i;
  char *args[8];
  char *md5coll = "./md5coll";
  char *md5_vk1_single = "./md5_vk1_single";
  char output1[64], output2[64], output3[64];

  struct option longopts[] = {
    {"help", 0, 0, 'h' },
    {"verbose", 0, 0, 'v' },
    {"badchars", 0, 0, 'b'},
    {"tunnel", 0, 0, 't' },
    {"seed", 1, 0, 's' },
    {0, 0, 0, 0}
  };

  progname = argv[0];
  tunnel = verbose = 0;
  while ( (ch = getopt_long(argc, argv, "bvhts:", longopts, &longind)) != -1) {
    switch (ch) {
    case 'h':
      usage();
      exit(0);
      break;
    case 'v':
      verbose = 1;
      break;
    case 'b':
      badchars = 1;
      break;
    case 't':
      tunnel = 1;
      break;
    case 's':
      snprintf(seedbuf, countof(seedbuf), "%08X", (unsigned int)strtoul(optarg, NULL, 16));
      seedbuf[countof(seedbuf)-1] = '\0';
      seed = seedbuf;
      break;
    default:
      usage();
      exit(2);
    }
  }

  argc -= optind;
  argv += optind;

  if (0 == argc) {
    die(0, "missing file argument");
    usage();
    exit(2);
  }

  filename = argv[0];

  /* check the file, size must be multipe of 64 */
  if (verbose) {
    printf("reading from file: %s\n", filename);
  }
  if ( stat(filename, &st) < 0) {
    die(1, "failed to stat file %s: %s", filename, strerror(errno));
  }
  len = st.st_size;
  if (verbose) {
    printf("file size: %d\n", len);
  }
  if (0 != (len % 64)) {
    die(1, "file %s has bad file size %d, should be multiple of 64", filename, len);
  }

  if ( (fd = open(filename, O_RDONLY)) < 0) {
    die(1, "failed to open file %s: %s", filename, strerror(errno));
  }

  if ( (addr = mmap(0, len, PROT_READ, MAP_PRIVATE, fd, 0)) == NULL) {
    die(1, "failed to mmap %d: %s", fd, strerror(errno));
  }
  if ( (data = (unsigned char *)malloc(len)) == NULL) {
    die(1, "failed to malloc %d bytes", len);
  }
  memcpy(data, addr, len);
  munmap(addr, len);

  if (verbose) {
    printf("calculating MD5 initial values\n");
  }
  MD5Init(&MD5c);
  MD5Update(&MD5c, data, len);
  snprintf(IV[0], countof(IV[0]), "0x%08X", (unsigned int) MD5c.buf[0]);
  snprintf(IV[1], countof(IV[1]), "0x%08X", (unsigned int) MD5c.buf[1]);
  snprintf(IV[2], countof(IV[2]), "0x%08X", (unsigned int) MD5c.buf[2]);
  snprintf(IV[3], countof(IV[3]), "0x%08X", (unsigned int) MD5c.buf[3]);
  IV[0][countof(IV[0])-1] = '\0';
  IV[1][countof(IV[1])-1] = '\0';
  IV[2][countof(IV[2])-1] = '\0';
  IV[3][countof(IV[3])-1] = '\0';
  MD5Final(junk, &MD5c);
  
  /* create args for program */
  if (tunnel) {
    /* check for local version */
    if (stat(md5_vk1_single, &st) < 0) {
      md5_vk1_single = md5_vk1_single + 2;
      if (verbose) {
	printf("assuming %s is in path\n", md5_vk1_single);
      }
    }
    args[0] = md5_vk1_single;
  } else {
    /* check for local version */
    if (stat(md5coll, &st) < 0) {
      md5coll = md5coll + 2;
      if (verbose) {
	printf("assuming %s is in path\n", md5coll);
      }
    }
    args[0] = md5coll;

  }
  /* keep looping until a collision is found with the desired values */
  do {
    i = 1;
    if (tunnel) {
      if (verbose) {
	printf("using seed: 0x%s\n", seed);
      }
      args[i++] = seed;
    }
    args[i++] = IV[0];
    args[i++] = IV[1];
    args[i++] = IV[2];
    args[i++] = IV[3];
    args[i++] = 0;
    if (verbose) {
      printf("using IV: {%s, %s, %s, %s}\n", IV[0], IV[1], IV[2], IV[3]);
      printf("launching %s\n", args[0]);
    }

    /* check for progs */
    if ( (pid = fork()) < 0) {
      die(1, "failed to fork: %s", strerror(errno));
    }
    if (pid) {
      if (verbose) {
	printf("waiting for child: %d\n", pid);
      }
      waitpid(pid, &status, 0);
      if (verbose) {
	printf("child returned: %d\n", status);
      }
      if (0 != status) {
	die(1, "failed when processing child program");
      }
    } else {
      /* launch child program */
      if (!verbose) {
	if ( (devnullfd = open("/dev/null", O_RDWR, 0666)) < 0) {
	  die(1, "failed to open /dev/null: %s", strerror(errno));
	}
	if (dup2(devnullfd, STDIN_FILENO) < 0) 
	  die(1, "failed to dup2(0): %s", strerror(errno));
	if (dup2(devnullfd, STDOUT_FILENO) < 0)
	  die(1, "failed to dup2(1): %s", strerror(errno));
	if (dup2(devnullfd, STDERR_FILENO) < 0)
	  die(1, "failed to dup2(2): %s", strerror(errno));
      }
      execvp(args[0], args);
      die(1, "execvp error: %s", strerror(errno));
    }

    if (tunnel) {
      snprintf(output1, countof(output1), "collision1_md5_%s.BIN", seed);
      snprintf(output2, countof(output2), "collision2_md5_%s.BIN", seed);
      snprintf(output3, countof(output3), "collision_md5_%s.TXT", seed);
      output1[countof(output1)-1] = '\0';
      output2[countof(output2)-1] = '\0';
      output3[countof(output3)-1] = '\0';
      if (check_files(data, len, output1, output2)) { 
	printf("found collisions using seed %s; file1=%s, file2=%s\n", seed, output1, output2);
	goto done;
      }
    } else {
      strcpy(output1, "m0.bin");
      strcpy(output2, "m1.bin");
      if (check_files(data, len, output1, output2)) { 
	printf("found collisions; file1=%s, file2=%s\n", output1, output2);
	goto done;
      }
    }
    /* remove failed files */
    if (unlink(output1) < 0) {
      printf("warning: failed to unlink %s: %s", output1, strerror(errno));
    }
    if (unlink(output2) < 0) {
      printf("warning: failed to unlink %s: %s", output2, strerror(errno));
    }
    if (tunnel) {
      if (unlink(output3) < 0) {
	printf("warning: failed to unlink %s: %s", output3, strerror(errno));
      }
    }

    if (tunnel) {
      /* need to create new seed */
      snprintf(seedbuf, countof(seedbuf), "%08X", (unsigned int)strtoul(seed, NULL, 16) + 1);
      seedbuf[countof(seedbuf)-1] = '\0';
      seed = seedbuf;
    }

  } while (1);

 done:
  free(data);
  exit(0);
}
  
